'''Unit tests for the MintingReward class. See unit_test_lease_calculations.ods
for a sense of where these numbers are coming from.
'''

from decimal import Decimal

from pyvsystems_rewards.address import Address
from pyvsystems_rewards.lease import Lease
from pyvsystems_rewards.minting_reward import MintingReward


def test_operation_fee():
    lease_one = Lease('lease_one_id', 'address_one', 3300000000, 1)
    lease_two = Lease('lease_two_id', 'address_one', 50000000000, 50)
    lease_three = Lease('lease_three_id', 'address_one', 50000000000, 125)
    active_leases = [lease_one, lease_two, lease_three]

    minting_reward = MintingReward('transaction_id', 3600000000, 3000, active_leases, Decimal('0.18'))

    assert minting_reward.operation_fee == 648000000


def test_interest():
    lease_one = Lease('lease_one_id', 'address_one', 3300000000, 1)
    lease_two = Lease('lease_two_id', 'address_one', 50000000000, 50)
    lease_three = Lease('lease_three_id', 'address_one', 50000000000, 125)
    active_leases = [lease_one, lease_two, lease_three]
    
    minting_reward = MintingReward('transaction_id', 3600000000, 3000, active_leases, Decimal('0.18'))

    assert minting_reward.interest == 2952000000


def test_interest_for_lease():
    lease_one = Lease('lease_one_id', 'address_one', 3300000000, 1)
    lease_two = Lease('lease_two_id', 'address_one', 50000000000, 50)
    lease_three = Lease('lease_three_id', 'address_one', 50000000000, 125)
    active_leases = [lease_one, lease_two, lease_three]
    
    minting_reward = MintingReward('transaction_id', 3600000000, 3000, active_leases, Decimal('0.18'))

    assert minting_reward.interest_for_lease(lease_one) == 97012713
    assert minting_reward.interest_for_lease(lease_two) == 1445873390
    assert minting_reward.interest_for_lease(lease_three) == 1409113897


def test_interest_for_address():
    lease_one = Lease('lease_one_id', 'address_one', 3300000000, 1)
    lease_two = Lease('lease_two_id', 'address_one', 50000000000, 50)
    lease_three = Lease('lease_three_id', 'address_two', 50000000000, 125)
    active_leases = [lease_one, lease_two, lease_three]

    minting_reward = MintingReward('transaction_id', 3600000000, 3000, active_leases, Decimal('0.18'))

    address = Address('address_one')
    address.start_lease(lease_one)
    address.start_lease(lease_two)

    assert minting_reward.interest_for_address(address) == 97012713 + 1445873390



from collections import OrderedDict

import requests

from .address import Address
from .lease import Lease
from .minting_reward import MintingReward
from .pool_distribution import PoolDistribution


class AddressFactory:
    TRANSACTION_CODE = {
        'Genesis': 1,
        'Payment': 2,
        'Lease': 3,
        'LeaseCancel': 4,
        'MintingTransaction': 5,
        'ContendSlotsTransaction': 6,
        'ReleaseSlotsTransaction': 7,
        'RegisterContractTransaction': 8,
        'ExecuteContractFunctionTransaction': 9,
        'DbPutTransaction': 10,
    }

    def __init__(
        self,
        api_url,
        hot_wallet_address,
        cold_wallet_address,
        operation_fee_percent
    ):
        self.total_interest = 0
        self.total_operation_fee = 0
        self.total_pool_distribution = 0
        self._api_url = api_url
        self._hot_wallet_address = hot_wallet_address
        self._cold_wallet_address = cold_wallet_address
        self._operation_fee_percent = operation_fee_percent
        self._addresses = None

        self.get_addresses()

    def get_addresses(self):
        if self._addresses is not None:
            return self._addresses.values()

        self._addresses = {}
        hot_wallet_transactions = self._get_transactions(self._hot_wallet_address)
        cold_wallet_transactions = self._get_transactions(self._cold_wallet_address)
        self._add_leases(hot_wallet_transactions)
        self._add_minting_rewards(cold_wallet_transactions)
        self._add_pool_distributions(cold_wallet_transactions)

        return self._addresses.values()

    def get_active_addresses(self, height):
        addresses = self.get_addresses()
        return [address for address in addresses if address.is_active(height)]

    def get_inactive_addresses(self, height):
        addresses = self.get_addresses()
        return [address for address in addresses if not address.is_active(height)]

    def _get_transactions(self, address):
        '''Returns all of the transactions associated with an address in
        increasing block height order.'''
        descending_transactions = OrderedDict()
        offset = 0
        LIMIT = 10000
        while True:
            transactions = requests.get(
                self._api_url + '/transactions/list',
                params={'address': address, 'limit': LIMIT, 'offset': offset}
            ).json();

            if transactions['size'] == 0:
                break

            for transaction in transactions['transactions']:
                descending_transactions[transaction['id']] = transaction

            offset += LIMIT

        ascending_transactions = []
        for transaction_id in reversed(descending_transactions):
            ascending_transactions.append(descending_transactions[transaction_id])

        return ascending_transactions

    def _add_leases(self, hot_wallet_transactions):
        for tx in hot_wallet_transactions:
            if tx['type'] == self.TRANSACTION_CODE['Lease']:
                address = tx['proofs'][0]['address']
                if address not in self._addresses:
                    self._addresses[address] = Address(address)

                address = self._addresses[address]
                address.start_lease(
                    Lease(tx['id'], address.address, tx['amount'], tx['height'])
                )

            elif tx['type'] == self.TRANSACTION_CODE['LeaseCancel']:
                address = tx['proofs'][0]['address']
                address = self._addresses[address]
                address.stop_lease(tx['leaseId'], tx['height'])

    def _get_active_leases(self, height):
        active_leases = []
        for address in self.get_addresses():
            active_leases.extend(address.active_leases(height))

        return active_leases

    def _add_minting_rewards(self, cold_wallet_transactions):
        for tx in cold_wallet_transactions:
            if tx['type'] == self.TRANSACTION_CODE['MintingTransaction']:
                minting_reward = MintingReward(
                    tx['id'],
                    tx['amount'],
                    tx['height'],
                    self._get_active_leases(tx['height']),
                    self._operation_fee_percent
                )
                self.total_interest += minting_reward.interest
                self.total_operation_fee += minting_reward.operation_fee

                for address in self.get_active_addresses(tx['height']):
                    address.add_minting_reward(minting_reward)

    def _add_pool_distributions(self, cold_wallet_transactions):
        for tx in cold_wallet_transactions:
            if (tx['type'] == self.TRANSACTION_CODE['Payment'] and
                    tx['proofs'][0]['address'] == self._cold_wallet_address):
                address = tx['recipient']
                if address not in self._addresses:
                    continue

                pool_distribution = PoolDistribution(
                    tx['id'],
                    address,
                    tx['amount'],
                    tx['fee'],
                    tx['height']
                )
                self.total_pool_distribution += pool_distribution.amount + pool_distribution.fee
                self._addresses[address].add_pool_distribution(pool_distribution)



from decimal import Decimal

class Lease:
    MAB_MATURES_AFTER_BLOCKS = 86400

    def __init__(self, lease_id, address, amount, start_height, stop_height=None):
        self.lease_id = lease_id
        self.address = address
        self.amount = amount
        self.start_height = start_height
        self.stop_height = stop_height
        self.total_interest = 0

    def is_active(self, height):
        if height < self.start_height:
            return False

        return self.stop_height is None or height < self.stop_height

    def get_mab(self, height):
        if not self.is_active(height):
            return None

        mab_modifier = Decimal(height - self.start_height) / Decimal(self.MAB_MATURES_AFTER_BLOCKS)
        mab_modifier = min(1, mab_modifier)
        return int(mab_modifier * self.amount)

    def add_minting_reward(self, minting_reward):
        self.total_interest += minting_reward.interest_for_lease(self)

    # The following functions are required to use the Lease as a key in
    # dictionaries.
    def __hash__(self):
        return hash(self.lease_id)

    def __eq__(self, other):
        return self.lease_id == other.lease_id

    def __ne__(self, other):
        return not(self == other)



from pyvsystems_rewards.address_factory import AddressFactory


def test_address_factory():
    # TODO this might be where we load in some real-world csv or json and
    # examine the output.
    factory = AddressFactory()



class Address:
    def __init__(self, address):
        self.address = address
        self._leases = {}
        self._pool_distributions = {}
        self.total_interest = 0
        self.total_pool_distribution = 0

    def leases(self):
        return self._leases.values()

    def start_lease(self, lease):
        self._leases[lease.lease_id] = lease

    def stop_lease(self, lease_id, height):
        lease = self._leases[lease_id]
        lease.stop_height = height

    def active_leases(self, height):
        return [l for l in self.leases() if l.is_active(height)]

    def inactive_leases(self, height):
        return [l for l in self.leases() if not l.is_active(height)]

    def is_active(self, height):
        return bool(self.active_leases(height))

    def add_minting_reward(self, minting_reward):
        self.total_interest += minting_reward.interest_for_address(self)
        for lease in self.active_leases(minting_reward.height):
            lease.add_minting_reward(minting_reward)

    def pool_distributions(self):
        return self._pool_distributions.values()

    def add_pool_distribution(self, pool_distribution):
        self._pool_distributions[pool_distribution.pool_distribution_id] = pool_distribution
        self.total_pool_distribution += pool_distribution.amount + pool_distribution.fee

    def total_interest_owed(self):
        return self.total_interest - self.total_pool_distribution



from pyvsystems_rewards.lease import Lease


def test_ongoing_lease_is_active():
    lease = Lease('lease_id', 'address', 100000000, 10)
    assert lease.is_active(9) == False
    assert lease.is_active(10) == True
    assert lease.is_active(11) == True

def test_closed_lease_is_active():
    lease = Lease('lease_id', 'address', 100000000, 10, 20)
    assert lease.is_active(9) == False
    assert lease.is_active(10) == True
    assert lease.is_active(19) == True
    assert lease.is_active(20) == False
    assert lease.is_active(21) == False

def test_get_mab():
    lease = Lease('lease_id', 'address', 100000000, 10, 10 + Lease.MAB_MATURES_AFTER_BLOCKS + 10)

    assert lease.get_mab(10 - 1) is None
    assert lease.get_mab(10) == 0
    assert lease.get_mab(10 + Lease.MAB_MATURES_AFTER_BLOCKS) == 100000000
    assert lease.get_mab(10 + Lease.MAB_MATURES_AFTER_BLOCKS + 1) == 100000000
    assert lease.get_mab(10 + Lease.MAB_MATURES_AFTER_BLOCKS + 10) is None



class PoolDistribution:
    def __init__(self, pool_distribution_id, address, amount, fee, height):
        self.pool_distribution_id = pool_distribution_id
        self.address = address
        self.amount = amount
        self.fee = fee
        self.height = height



from decimal import Decimal

from pyvsystems_rewards.address import Address
from pyvsystems_rewards.lease import Lease
from pyvsystems_rewards.minting_reward import MintingReward


def test_add_minting_reward():
    '''Check that adding a minting reward to an address increments totals.'''
    lease_one = Lease('lease_one_id', 'address_one', 3300000000, 1)
    lease_two = Lease('lease_two_id', 'address_one', 50000000000, 50)
    lease_three = Lease('lease_three_id', 'address_two', 50000000000, 125)
    active_leases = [lease_one, lease_two, lease_three]

    minting_reward = MintingReward('transaction_id', 3600000000, 3000, active_leases, Decimal('0.18'))

    address = Address('address_one')
    address.start_lease(lease_one)
    address.start_lease(lease_two)

    assert address.total_interest == 0
    address.add_minting_reward(minting_reward)
    assert address.total_interest == 97012713 + 1445873390



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='pyvsystems_rewards',
    version='0.0.3',
    author='Chapman Shoop',
    author_email='chapman.shoop@gmail.com',
    description='A library for calculating v.systems supernode rewards',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/belovachap/pyvsystems_rewards',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



from decimal import Decimal


class MintingReward:
    def __init__(
        self,
        minting_reward_id,
        amount,
        height,
        active_leases,
        operation_fee_percent
    ):
        self.minting_reward_id = minting_reward_id
        self.amount = amount
        self.height = height
        self.operation_fee = int(amount * operation_fee_percent)
        self.interest = amount - self.operation_fee

        # Stage one of distribution.
        total_mab = sum([lease.get_mab(height) for lease in active_leases])
        self._lease_to_interest = {}
        stage_one_interest = 0
        for lease in active_leases:
            i = int(self.interest * (Decimal(lease.get_mab(height)) / Decimal(total_mab)))
            stage_one_interest += i
            self._lease_to_interest[lease] = i

        # Stage two of distribution.
        remaining_interest = self.interest - stage_one_interest
        ordered_leases = sorted(active_leases, key=lambda lease: lease.lease_id)
        ordered_leases = sorted(ordered_leases, key=lambda lease: lease.start_height)
        for lease in ordered_leases:
            if remaining_interest <= 0:
                break

            self._lease_to_interest[lease] += 1
            remaining_interest -= 1

    def interest_for_lease(self, lease):
        return self._lease_to_interest[lease]

    def interest_for_address(self, address):
        interest = 0
        for lease in self._lease_to_interest:
            if lease.address == address.address:
                interest += self.interest_for_lease(lease)

        return interest
